from collections import defaultdict
import numpy as np
import pandas as pd
from scipy.stats import betabinom as sp_betabinom

#%%

# logger = logging.getLogger(__name__)

#%%


def split(strng, sep, pos):
    strng = strng.split(sep)
    return sep.join(strng[:pos]), sep.join(strng[pos:])


def get_number_of_lines(filename):
    with open(filename, "r") as f:
        counter = 0
        for _ in f:
            counter += 1
    return counter


def read_filename_lca(file_lca):

    N_lines = get_number_of_lines(file_lca)

    d_tax_id_to_read_ids = defaultdict(set)
    with open(file_lca, "r") as f:

        for irow, row in enumerate(f):

            if irow == 0:
                continue

            read_id, rest = split(row.strip(), sep=":", pos=7)
            _, lca = split(rest, sep="\t", pos=1)
            tax_id = int(lca.split(":")[0])
            d_tax_id_to_read_ids[tax_id].add(read_id)

    d_tax_id_to_read_ids = dict(d_tax_id_to_read_ids)

    for key, val in d_tax_id_to_read_ids.items():
        d_tax_id_to_read_ids[key] = list(val)

    return d_tax_id_to_read_ids


def compute_df_mismatches_wide(df_mismatches):

    max_pos = df_mismatches.position.max()

    d_direction = {
        "forward": {
            "query": "position > 0",
            "symbol": "+",
        },
        "reverse": {
            "query": "position < 0",
            "symbol": "-",
        },
    }

    df_mismatches_wide = []
    for direction in ["forward", "reverse"]:
        for variable in ["k", "N", "f"]:
            col_names = [
                f"{variable}{d_direction[direction]['symbol']}{i}"
                for i in range(1, max_pos + 1)
            ]
            columns = {i + 1: col for i, col in enumerate(col_names)}

            df_mismatches_wide.append(
                df_mismatches.query(d_direction[direction]["query"])
                .pivot(index="tax_id", columns="|z|", values=variable)
                .rename(columns=columns)
            )

    df_mismatches_wide = pd.concat(df_mismatches_wide, axis="columns")

    return df_mismatches_wide


def merge(
    df_mismatches,
    df_fit_results,
):

    # merge the mismatches counts (as a wide dataframe) into the fit results dataframe
    df_mismatches_wide = compute_df_mismatches_wide(df_mismatches)
    df_results = pd.merge(df_fit_results, df_mismatches_wide, on=["tax_id"])

    columns_order = [
        "tax_id",
        "tax_name",
        "tax_rank",
        "sample",
        "N_reads",
        "N_alignments",
        #
        "lambda_LR",
        "D_max",
        "mean_L",
        "mean_GC",
        "q",
        "A",
        "c",
        "phi",
        "rho_Ac",
        "valid",
        "asymmetry",
    ]

    columns_order += [col for col in df_fit_results.columns if not col in columns_order]
    columns_order += list(df_mismatches_wide.columns)

    df_results = df_results[columns_order]

    return df_results


def get_database_read_ids(config):
    read_ids_mapping = read_filename_lca(config["path_lca"])
    return read_ids_mapping

    # db = database(filename_database)
    # db.save(d_tax_id_to_read_ids)



import pandas as pd
from pathlib import Path
import subprocess
import shlex
import shutil
from logger_tt import logger
from multiprocessing import current_process
import json
from metaDMG.fit import mismatches, fits, results


#%%


def do_run(targets, forced=False):

    if forced:
        return True

    if not isinstance(targets, list):
        targets = [targets]

    if all(Path(target).exists() for target in targets):
        return False
    else:
        return True


def data_dir(config, name, suffix="parquet"):
    target = Path(config["dir"]) / name / f"{config['sample']}.{name}.{suffix}"
    return str(target)


#%%


def get_LCA_command(config):
    sample = config["sample"]
    bam = config["samples"][sample]
    lca_rank = f"-lca_rank {config['lca_rank']}" if config["lca_rank"] != "" else ""

    command = (
        f"{config['metaDMG-lca']} lca "
        f"-bam {bam} "
        f"-outnames {sample} "
        f"-names {config['names']} "
        f"-nodes {config['nodes']} "
        f"-acc2tax {config['acc2tax']} "
        f"-simscorelow {config['simscorelow']} "
        f"-simscorehigh {config['simscorehigh']} "
        f"-editdistmin {config['editdistmin']} "
        f"-editdistmax {config['editdistmax']} "
        f"{lca_rank} "
        f"-minmapq {config['minmapq']} "
        f"-howmany {config['max_position']} "
        f"-weighttype {config['weighttype']} "
        f"-fix_ncbi {config['fix_ncbi']} "
    )
    return command


def get_LCA_mismatches_command(config):
    sample = config["sample"]
    bdamage = f"{sample}.bdamage.gz"
    lca_stat = f"{sample}.lca.stat"

    command = (
        f"{config['metaDMG-lca']} print_ugly "
        f"{bdamage} "
        f"-names {config['names']} "
        f"-nodes {config['nodes']} "
        f"-lcastat {lca_stat} "
    )
    return command


def run_command(command):

    p = subprocess.Popen(
        shlex.split(command),
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    for line in iter(p.stdout.readline, b""):
        if line:
            line = line.decode("utf-8")
            if line.endswith("\n"):
                line = line[:-1]
            yield line


def move_files(config):
    sample = config["sample"]

    d_move_source_target = {
        f"{sample}.bdamage.gz.uglyprint.mismatch.txt": config["path_mismatches_txt"],
        f"{sample}.bdamage.gz.uglyprint.stat.txt": config["path_mismatches_stat"],
        f"{sample}.lca": config["path_lca"],
    }
    for source_path, target_path in d_move_source_target.items():
        logger.debug(f"Moving {source_path} to {target_path}.")
        Path(target_path).parent.mkdir(parents=True, exist_ok=True)
        shutil.move(source_path, target_path)


def delete_files(config):
    sample = config["sample"]

    bam = Path(config["samples"][sample]).stem  # .name

    paths_to_remove = [
        f"{sample}.lca.stat",
        f"{sample}.bdamage.gz",
        f"{sample}.log",
        *list(Path(".").glob(f"*{bam}*.bam.bin")),
    ]
    for path in paths_to_remove:
        logger.debug(f"Removing {path}.")
        Path(path).unlink()


def run_LCA(config, forced=False):

    logger.info(f"Getting LCA.")

    targets = [
        config["path_mismatches_txt"],
        config["path_mismatches_stat"],
        config["path_lca"],
    ]

    if do_run(targets, forced=forced):

        logger.info(f"LCA has to be computed. " "This can take a while, please wait.")

        command_LCA = get_LCA_command(config)
        command_LCA_mismatches = get_LCA_mismatches_command(config)

        for line in run_command(command_LCA):
            logger.debug(line)
        for line in run_command(command_LCA_mismatches):
            logger.debug(line)

        move_files(config)
        delete_files(config)

    else:
        logger.info(f"LCA already been run before.")


#%%


def get_df_mismatches(config, forced=False):

    logger.info(f"Getting df_mismatches")

    target = data_dir(config, name="mismatches")

    if do_run(target, forced=forced):
        logger.info(f"Computing df_mismatches.")
        df_mismatches = mismatches.compute(config)
        Path(target).parent.mkdir(parents=True, exist_ok=True)
        df_mismatches.to_parquet(target)

    else:
        logger.info(f"Loading df_mismatches.")
        df_mismatches = pd.read_parquet(target)

    return df_mismatches


#%%


def get_df_fit_results(config, df_mismatches, forced=False):

    logger.info(f"Getting df_fit_results.")

    target = data_dir(config, name="fit_results")
    if do_run(target, forced=forced):
        logger.info(f"Computing df_fit_results")
        df_fit_results = fits.compute(config, df_mismatches)
        Path(target).parent.mkdir(parents=True, exist_ok=True)
        df_fit_results.to_parquet(target)

    else:
        logger.info(f"Loading df_fit_results.")
        df_fit_results = pd.read_parquet(target)

    return df_fit_results


#%%


def get_df_results(config, df_mismatches, df_fit_results, forced=False):

    logger.info(f"Getting df_results.")

    target = data_dir(config, name="results")

    if do_run(target, forced=forced):
        logger.info(f"Computing df_results.")
        df_results = results.merge(
            df_mismatches,
            df_fit_results,
        )
        Path(target).parent.mkdir(parents=True, exist_ok=True)
        df_results.to_parquet(target)

    else:
        logger.info(f"Loading df_results")
        df_results = pd.read_parquet(target)

    return df_results


#%%


class DB:
    def __init__(self, path):
        self.path = path

    def save(self, d):
        Path(self.path).parent.mkdir(parents=True, exist_ok=True)
        with open(self.path, "w") as write_file:
            json.dump(d, write_file, indent=4)

    def load(self):
        with open(self.path, "r") as read_file:
            d = json.load(read_file)
        return {int(k): v for k, v in d.items()}


def get_database_read_ids(config, forced=False):

    logger.info(f"Getting read_ids_mapping.")

    target = data_dir(config, name="database", suffix="json")

    database = DB(target)

    if do_run(target, forced=forced):
        logger.info(f"Computing read_ids_mapping.")
        read_ids_mapping = results.get_database_read_ids(config)
        database.save(read_ids_mapping)

    else:
        logger.info(f"Loading read_ids_mapping.")
        read_ids_mapping = database.load()

    return read_ids_mapping


#%%


def run_single_config(config):

    current_process().name = config["sample"]

    run_LCA(config)
    df_mismatches = get_df_mismatches(config)
    df_fit_results = get_df_fit_results(config, df_mismatches)
    df_results = get_df_results(config, df_mismatches, df_fit_results, forced=False)
    read_ids_mapping = get_database_read_ids(config)

    logger.info("Finished.")

    return df_mismatches, df_fit_results, df_results



from logger_tt import setup_logging, logger
from datetime import datetime
from importlib import resources


def setup_logger():

    now = datetime.now()
    now_str = now.strftime("%Y-%m-%d__%H-%M")
    log_path = f"logs/log__{now_str}.log"

    with resources.path("metaDMG.loggers", "log_config.yaml") as p:
        config_path = p

    setup_logging(
        config_path=config_path,
        log_path=log_path,
    )
